/* Routines for performing Temporary Expression Replacement (TER) in SSA trees.
   Copyright (C) 2003-2022 Free Software Foundation, Inc.
   Contributed by Andrew MacLeod  <amacleod@redhat.com>

This file is part of GCC.

GCC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3, or (at your option)
any later version.

GCC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING3.  If not see
<http://www.gnu.org/licenses/>.  */

#include "config.h"
#include "system.h"
#include "coretypes.h"
#include "backend.h"
#include "tree.h"
#include "gimple.h"
#include "predict.h"
#include "tree-pass.h"
#include "ssa.h"
#include "cgraph.h"
#include "fold-const.h"
#include "stor-layout.h"
#include "gimple-iterator.h"
#include "gimple-walk.h"
#include "tree-ssa-loop-manip.h"
#include "tree-ssa-loop-niter.h"
#include "tree-cfg.h"
#include "cfgloop.h"
#include "tree-vectorizer.h"
#include "tree-ssa-propagate.h"
#include "dbgcnt.h"
#include "tree-scalar-evolution.h"
#include "stringpool.h"
#include "attribs.h"
#include "gimple-pretty-print.h"
#include "opt-problem.h"
#include "internal-fn.h"
#include "tree-ssa-sccvn.h"
#include "vec-perm-indices.h"

// Operation1:
// If a VEC_PERM_EXPR has all uses as a permute-like expression. Change all
// the use into a new VEC_PERM_EXPR, and delete the old one.
// And VEC_PERM_EXPR can use 255 as a number in operand3 which means to put
// a 0 in the dest operand. As a result, we can turn VEC_UNPACK_LO/HI_EXPR
// into a VEC_PERM_EXPR.
//
// Case 1:
// vect__2 = VEC_PERM_EXPR <vect__0, vect__1, { 0, 1, 2, 4, 5, 6, 8, 9 }>;
// vect__3 = [vec_unpack_lo_expr] vect__2;
// vect__4 = [vec_unpack_hi_expr] vect__2;
// ==>
// vect__3 = VEC_PERM_EXPR <vect__0, vect__0, { 0, 255, 1, 255, 2, 255, 4, 255 }>;
// vect__4 = VEC_PERM_EXPR <vect__0, vect__1, { 5, 255, 6, 255, 8, 255, 9, 255 }>;
//
// Case 2:
// vect__1 = VEC_PERM_EXPR <vect__0, vect__0, { 0, 1 }>;
// vect__2 = VEC_PERM_EXPR <vect__1, vect__1, { 0, 0 }>;
// vect__3 = VEC_PERM_EXPR <vect__1, vect__1, { 1, 1 }>;
// ==>
// vect__2 = VEC_PERM_EXPR <vect__0, vect__0, { 0, 0 }>;
// vect__3 = VEC_PERM_EXPR <vect__0, vect__0, { 1, 1 }>;
//
// Operation2:
// While a VEC_PERM_EXPR has different src, and it's use only use one. It
// can be performed either.
//
// Case:
// vect__2 = VEC_PERM_EXPR <vect__0, vect__1, { 5, 6, 8, 9 }>;
// vect__3 = [vec_unpack_float_lo_expr] vect__2;
// vect__4 = [vec_unpack_float_hi_expr] vect__2;
// ==>
// _0 = VEC_PERM_EXPR <vect__0, vect__0, { 5, 6, 255, 255 }>;
// vect__3 = [vec_unpack_float_lo_expr] _0;
// _1 = VEC_PERM_EXPR <vect__1, vect__1, { 0, 1, 255, 255 }>;
// vect__4 = [vec_unpack_float_lo_expr] _1;
void
permute_stmt_operation(gimple *stmt)
{
  imm_use_iterator iter;
  gimple *use_stmt;
  unsigned int perm[256],n;

  n=VECTOR_CST_NELTS (gimple_assign_rhs3(stmt)).to_constant ();

  for (unsigned int i=0; i<n; i++)
    perm[i] = TREE_INT_CST_ELT(VECTOR_CST_ELT(gimple_assign_rhs3(stmt),i), 0);
  tree rtype = copy_node (unsigned_type_node);
  tree lhs = gimple_assign_lhs (stmt);
  unsigned int permnew[256];
  FOR_EACH_IMM_USE_STMT (use_stmt, iter, lhs)
    {
      for (unsigned int i=0; i<n; i++)
	permnew[i] = 255;
      switch (gimple_assign_rhs_code(use_stmt))
        {
	  case VEC_UNPACK_LO_EXPR:
            for (unsigned int i=0; i<n/2; i++)
	      permnew[i*2] = perm[i];
	    break;

          case VEC_UNPACK_HI_EXPR:
            for (unsigned int i=0; i<n/2; i++)
	      permnew[i*2] = perm[i+n/2];
	    break;

	  case VEC_UNPACK_FLOAT_LO_EXPR:
            for (unsigned int i=0; i<n/2; i++)
	      permnew[i] = perm[i];
	    break;

	  case VEC_UNPACK_FLOAT_HI_EXPR:
            for (unsigned int i=0; i<n/2; i++)
	      permnew[i] = perm[i+n/2];
	    break;

	  case VEC_PERM_EXPR:
	    for (unsigned int i=0; i<n; i++ )
	      permnew[i] = perm[TREE_INT_CST_ELT(
			 VECTOR_CST_ELT(gimple_assign_rhs3(use_stmt),i), 0)];
	    break;

          default:
	    gcc_unreachable ();
	}
      	if (tree_to_uhwi(TYPE_SIZE(TREE_TYPE(gimple_assign_lhs(stmt))))>=256)
	  {
	    /* vpshufb for ymm only works intra lanes, it is not
	       possible to shuffle bytes in between the lanes.  */
	    for (unsigned int i = 0; i < n; ++i)
	      if ((permnew[i]!=255) || ((permnew[i] ^ i) & (n / 2)))
		{
                  if (dump_enabled_p ())
                    dump_printf_loc (MSG_MISSED_OPTIMIZATION, vect_location,
		          "Expr for ymm should match only one insn.\n"
		         );
		  return;
		}
	  }

      tree char_vectype = TREE_TYPE(gimple_assign_rhs3(stmt));
      tree op0 = gimple_assign_rhs1(stmt);
      tree op1 = gimple_assign_rhs2(stmt);
      unsigned int t=0;
      for (unsigned int i=0; i<n; i++)
	if (permnew[i] != 255)
	  t |= permnew[i]<n?1:2;
      if (t == 1) op1 = op0;
      if (t == 2)
        {
	  op0 = op1;
	  for(unsigned int i=0;i<n;i++)
            if (permnew[i] != 255)
	      permnew[i] &= n-1;
        }

      vec_perm_builder elts (n, n, 1);
      for (unsigned i = 0; i < n; i++)
        elts.quick_push (permnew[i]);
      vec_perm_indices indices (elts, 1, 256);
      tree bswap_vconst = vec_perm_indices_to_tree (char_vectype, indices);
      gimple_stmt_iterator gsi;
      tree new_dest,new_temp;
      gimple *new_stmt,*new_use;
      switch (gimple_assign_rhs_code(use_stmt))
	{
	  case VEC_UNPACK_LO_EXPR:
	  case VEC_UNPACK_HI_EXPR:
	  case VEC_PERM_EXPR:
            new_dest = gimple_assign_lhs(use_stmt);
            new_stmt = gimple_build_assign (new_dest, VEC_PERM_EXPR,
			   op0, op1, bswap_vconst);
            if (dump_enabled_p ())
              dump_printf_loc (MSG_NOTE, vect_location,
		               "New stmt: %G",
		               new_stmt);
            gsi = gsi_for_stmt (use_stmt);
            gsi_replace(&gsi, new_stmt, true);
	    break;

	  case VEC_UNPACK_FLOAT_LO_EXPR:
	  case VEC_UNPACK_FLOAT_HI_EXPR:
            new_dest = gimple_assign_lhs(use_stmt);
	    new_temp = make_ssa_name (TREE_TYPE (lhs));
            new_stmt = gimple_build_assign (new_temp, VEC_PERM_EXPR,
			   op0, op1, bswap_vconst);
            if (dump_enabled_p ())
              dump_printf_loc (MSG_NOTE, vect_location,
		               "New stmt: %G",
		               new_stmt);
            gsi = gsi_for_stmt (use_stmt);
	    gsi_insert_before(&gsi, new_stmt, GSI_NEW_STMT);
            new_use = gimple_build_assign (new_dest,
			   VEC_UNPACK_FLOAT_LO_EXPR,
			   new_temp);
            if (dump_enabled_p ())
              dump_printf_loc (MSG_NOTE, vect_location,
		               "New stmt: %G",
		               new_use);
            gsi = gsi_for_stmt (use_stmt);
            gsi_replace(&gsi, new_use, true);
	    break;

	  default:
	   gcc_unreachable ();
        }
    }
}

// Analysis all the gimple and find the match case.
bool
permute_stmt_analysis(basic_block bb, gimple *stmt)
{
  imm_use_iterator iter;
  gimple *use_stmt;

  if (gimple_code(stmt) != GIMPLE_ASSIGN
      || gimple_assign_rhs_code(stmt) != VEC_PERM_EXPR) return false;
  if (dump_enabled_p ())
    dump_printf_loc (MSG_NOTE, vect_location,
		       "\nFind VEC_PERM_EXPR: %G\nUnsigned_p=%d\n",stmt,
		     TYPE_UNSIGNED(TREE_TYPE(gimple_assign_lhs(stmt))));

  tree lhs = gimple_assign_lhs (stmt);
  FOR_EACH_IMM_USE_STMT (use_stmt, iter, lhs)
    {
      if (dump_enabled_p ())
        dump_printf_loc (MSG_NOTE, vect_location,
		          "Find user: %G",
		         use_stmt);
      if (gimple_bb(use_stmt) != bb) {
            if (dump_enabled_p ())
     		  dump_printf_loc (MSG_MISSED_OPTIMIZATION, vect_location,
                             "Shuffle fusion failed: use not in the same bb.\n"
		         );
	      return false;
      }
      if (gimple_code(use_stmt) == GIMPLE_ASSIGN
	  && gimple_assign_rhs_code(use_stmt) == VEC_PERM_EXPR
	  && gimple_assign_rhs1(use_stmt) == gimple_assign_rhs2(use_stmt)
	  && TREE_TYPE(gimple_assign_lhs(stmt))
	     == TREE_TYPE(gimple_assign_rhs1(stmt)))
	continue;

      if (gimple_code(use_stmt)==GIMPLE_ASSIGN
	  && (gimple_assign_rhs_code(use_stmt) == VEC_UNPACK_LO_EXPR
	   || gimple_assign_rhs_code(use_stmt) == VEC_UNPACK_HI_EXPR)
	  && TYPE_SIGN(TREE_TYPE(gimple_assign_lhs(stmt))) == UNSIGNED)
        continue;

      if (dump_enabled_p ())
        dump_printf_loc (MSG_MISSED_OPTIMIZATION, vect_location,
                         "Shuffle fusion failed: use is not match.\n"
   	         	);
      return false;
    }
  if (dump_enabled_p ())
    dump_printf_loc (MSG_NOTE, vect_location,
		          "Analysis True.\n"
		         );
  return true;
}

bool
permute_stmt_analysis2(basic_block bb, gimple *stmt)
{
  imm_use_iterator iter;
  gimple *use_stmt;
  unsigned int perm[256], n, tl=0, th=0;

  if (gimple_code(stmt) != GIMPLE_ASSIGN
      || gimple_assign_rhs_code(stmt) != VEC_PERM_EXPR) return false;
  if (gimple_assign_rhs1(stmt)==gimple_assign_rhs2(stmt))
    {
      if (dump_enabled_p ())
        dump_printf_loc (MSG_MISSED_OPTIMIZATION, vect_location,
		          "Shuffle fusion2 failed: expr has the same rhs.\n"
		         );
      return false;
    }
  tree lhs = gimple_assign_lhs (stmt);
  n=VECTOR_CST_NELTS (gimple_assign_rhs3(stmt)).to_constant ();
  if (n>32) 
    {
      if (dump_enabled_p ())
        dump_printf_loc (MSG_MISSED_OPTIMIZATION, vect_location,
		          "Shuffle fusion2 failed: expr has too many nelts.\n"
		         );
      return false;
    }

  for (unsigned int i=0; i<n; i++)
    perm[i] = TREE_INT_CST_ELT(VECTOR_CST_ELT(gimple_assign_rhs3(stmt),i), 0);
  for (unsigned int i=0; i<n/2; i++)
    if (perm[i]!=255)
      tl |= perm[i]<n?1:2;
  for (unsigned int i=n/2; i<n; i++)
    if (perm[i]!=255)
      th |= perm[i]<n?1:2;

  FOR_EACH_IMM_USE_STMT (use_stmt, iter, lhs)
    {
      if (gimple_bb(use_stmt) != bb)
	      return false;
      if (gimple_code(use_stmt) != GIMPLE_ASSIGN
	  || ((gimple_assign_rhs_code(use_stmt) != VEC_UNPACK_FLOAT_LO_EXPR
	       || tl == 3)
	  && (gimple_assign_rhs_code(use_stmt) != VEC_UNPACK_FLOAT_HI_EXPR
	       || th == 3)))
	{
          if (dump_enabled_p ())
            dump_printf_loc (MSG_MISSED_OPTIMIZATION, vect_location,
		          "Shuffle fusion2 failed: user need both rhs.\n"
		         );
	  return false;
	}

    }
  if (dump_enabled_p ())
    dump_printf_loc (MSG_NOTE, vect_location,
		          "Act 2 Analysis True.\n"
		         );
  return true;
}

void
permute_bb_analysis(function *fun)
{
  basic_block bb;

  FOR_EACH_BB_FN (bb, fun)
    {
      for (gimple_stmt_iterator gsi = gsi_start_bb (bb); !gsi_end_p (gsi);
           gsi_next (&gsi))
        {
  	  gimple *stmt = gsi_stmt (gsi);
	  if (permute_stmt_analysis(bb, stmt) ||
	      permute_stmt_analysis2(bb, stmt))
	    permute_stmt_operation(stmt);
        }
    }
}

/*  Entry point to shuffle fusion phase.  */

namespace {

const pass_data pass_data_shuffle_fusion =
{
  GIMPLE_PASS, /* type */
  "shuffle-fusion", /* name */
  OPTGROUP_VEC, /* optinfo_flags */
  TV_SHUFFLE_FUSION, /* tv_id */
  ( PROP_ssa | PROP_cfg ), /* properties_required */
  0, /* properties_provided */
  0, /* properties_destroyed */
  0, /* todo_flags_start */
  TODO_update_ssa, /* todo_flags_finish */
};

class pass_shuffle_fusion : public gimple_opt_pass
{
public:
  pass_shuffle_fusion (gcc::context *ctxt)
    : gimple_opt_pass (pass_data_shuffle_fusion, ctxt)
  {}

  /* opt_pass methods: */
  virtual bool gate (function *) { return flag_shuffle_fusion; }
  virtual unsigned int execute (function *);

}; // class pass_slp_vectorize


unsigned int
pass_shuffle_fusion::execute (function *fun)
{
  auto_purge_vect_location sentinel;
  basic_block bb;

  bool in_loop_pipeline = scev_initialized_p ();
  if (!in_loop_pipeline)
    {
      loop_optimizer_init (LOOPS_NORMAL);
      scev_initialize ();
    }

  /* Mark all stmts as not belonging to the current region and unvisited.  */
  FOR_EACH_BB_FN (bb, fun)
    {
      for (gphi_iterator gsi = gsi_start_phis (bb); !gsi_end_p (gsi);
	   gsi_next (&gsi))
	{
	  gphi *stmt = gsi.phi ();
	  gimple_set_uid (stmt, -1);
	  gimple_set_visited (stmt, false);
	}
      for (gimple_stmt_iterator gsi = gsi_start_bb (bb); !gsi_end_p (gsi);
	   gsi_next (&gsi))
	{
	  gimple *stmt = gsi_stmt (gsi);
	  gimple_set_uid (stmt, -1);
	  gimple_set_visited (stmt, false);
	}
    }

  permute_bb_analysis (fun);

  if (!in_loop_pipeline)
    {
      scev_finalize ();
      loop_optimizer_finalize ();
    }

  return 0;
}

} // anon namespace

gimple_opt_pass *
make_pass_shuffle_fusion (gcc::context *ctxt)
{
  return new pass_shuffle_fusion (ctxt);
}
